#!/usr/local/bin/python2.6
#
# WadcomBlog - wadcom's Blog engine
#
# Copyright (c) 2006-2008 Vlad Skvortsov, vss@73rus.com
#
# $Id: wadcomblog.py 1823 2008-07-05 05:27:18Z vss $
# $HeadURL: https://svn.xiolabs.com/vss/wadcomblog/tags/0.3/src/wadcomblog.py $
#
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, 
# this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright notice, 
# this list of conditions and the following disclaimer in the documentation 
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#

import ConfigParser
import datetime
import email
import errno
import os
import shutil
import sys
import time

import lib.ezt

# Version information
VERSION = "0.3"
revision = "???"

DIRECTIVE = "<!-- WADCOMBLOG:"
DIRECTIVE_HEADER_END = DIRECTIVE + " header-end"

def makedirs(dirs):
    try:
        os.makedirs(dirs)
    except OSError, (code, str):
        if code != errno.EEXIST:
            raise

class _data_obj:
    def __init__(self, d):
        vars(self).update(d)

class BlogEntry:
    class TimeZone(datetime.tzinfo):
        def __init__(self):
            pass

        def dst(self, dt):
            return None

        def tzname(self, dt):
            return time.tzname[0]

    def __init__(self, cfg, fname):

        self.cfg = cfg
        self.name = os.path.split(fname)[-1]

	# XXX: source file name hardcoded as "index.html"
        name = os.path.join(cfg.src, fname, "index.html")
        data = email.message_from_file(open(name))

        if "Title" not in data: 
            sys.stderr.write("%s: no 'Title' header\n" % name)
            raise ValueError

        self.title = data["Title"]

        if "Timestamp" not in data: 
            sys.stderr.write("%s: no 'Timestamp' header\n" % name)
            raise ValueError

        self.datetime = data["Timestamp"]

        # XXX: day/month names are subject to locale settings!
        self.rfcdatetime = datetime.datetime(
            int(self.datetime[0:4]), int(self.datetime[4:6]),
            int(self.datetime[6:8]), int(self.datetime[8:10]),
            int(self.datetime[10:12]),
            0, # seconds,
            0, # microseconds,
            self.TimeZone()
        ).strftime("%a, %d %b %Y %T %Z")

        tags = []

        if "Tags" in data: 
            tags = data["Tags"].strip().split()
            tags.sort()

        self.tags = tags

        # User-defined headers
        self.user_header = {}

        for h in cfg.mandatory_headers:
            if h not in data:
                sys.stderr.write("%s: no mandatory user header '%s'\n" %
                    (name, h)
                )
                raise ValueError

            self.user_header[h.lower()] = data[h]

        # Now turn user-defined headers into a template object
        self.user_header = _data_obj(self.user_header)

        self.header = self.body = None

        contents = ""
        for str in data.get_payload().split("\n"):
            
            if str.strip().startswith(DIRECTIVE_HEADER_END):

                if self.header is None:

                    self.header = contents
                    contents = ""

                else:
                    sys.stderr.write(
                        "%s: duplicate 'header-end' directive, ignored\n"
                        % name
                    )

            else:
                contents += "\n" + str

        if self.header is None:   
            sys.stderr.write("%s: no 'header-end' directive\n" % name)
            raise ValueError

        self.body = contents

        # A list of directory components leading to the article.
        self.location = [
            self.datetime[0:4], self.datetime[4:6], self.datetime[6:8],
            self.name
        ]

    def install(self, dst):
        entry_path = self.path(dst)
        makedirs(entry_path)
        install_page(self.cfg, cfg.template["entry"], entry_path,
            len(self.location), "", 
            lambda root: prepare_entry(cfg, root, entry)
        )

    def path(self, root):
        if root is not None:
            return os.path.join(root, *self.location)
        else:
            return None

class Tag:
    def __init__(self, name, entry):
        self.entries = [ entry ]
        self.name = name

    def add_entry(self, entry):
        self.entries.append(entry)

class TagCloud:
    def __getitem__(self, key):
        return self.tags[key]

    def __init__(self):
        self.tags = {}

    def __iter__(self):
        tags = self.tags.keys()
        tags.sort()
        return tags.__iter__()

    def add_entry(self, entry):
        for tag in entry.tags:
            if tag in self.tags:
                self.tags[tag].add_entry(entry)
            else:
                self.tags[tag] = Tag(tag, entry)

    def rank(self, maxrating):
        """
        Go over all exisiting tags and assign rating from 1 (most popular)
        to MAXRATING (least popular).
        """

        assert maxrating >= 1

        lens = {}
        for l in map(lambda x: len(self.tags[x].entries), self.tags.keys()):
            lens[l] = True

        lens = lens.keys()
        lens.sort()

        # Number of different entry-per-tag counts
        weights = len(lens)

        if weights < maxrating:
            maxrating = weights
            assert maxrating >= 1

        step = float(weights) / maxrating

        weight = {}
        next = step
        rank = maxrating
        for i in range(weights):
            if i >= next:
                next += step
                rank -= 1

            weight[lens[i]] = rank

        for tag in self.tags.keys():
            self.tags[tag].rank = weight[len(self.tags[tag].entries)]

class Config:
    def __init__(self, fname):

        s = os.path.split(fname)
        if s[0]:
            self.src = s[0]
        else:
            self.src = "."

        parser = ConfigParser.ConfigParser()
        parser.read(fname)

        self.template = self.root_url = self.blog_title = self.feed_fname = \
            self.feed_size = None

        self.entries = []

        # Fetch data from the configuration
        self.blog_title = parser.get("general", "blog-title")
        self.front_page_caption = parser.get("general", "front-page-caption")
        self.entries_per_page = int(parser.get("general", "entries-per-page"))
        self.feed_fname = parser.get("general", "feed")
        self.feed_size = int(parser.get("general", "feed-size"))
        self.index_first_fname = parser.get(
	    "general", "index-first-fname", raw=True
	)
        self.index_rest_fname = parser.get(
	    "general", "index-rest-fname", raw=True
	)

        try:
            yesno = parser.get(
                "general", 
                "append-first-index-fname-to-links"
            ).lower()
        except ConfigParser.NoOptionError:
            yesno = "no"

        self.append_first_index_fname_to_links = (yesno == "yes")

        self.root_url = parser.get("general", "root")
        self.static = parser.get("general", "static").split()
        self.tag_ranks = int(parser.get("general", "tag-ranks"))

        if self.feed_size < 1:
            sys.stderr.write(
                "Invalid general/feed-size value: %s\n" % self.feed_size
            )
            sys.exit(1)

        try:
            self.mandatory_headers = \
                parser.get("general", "mandatory-headers").split()
        except ConfigParser.NoOptionError:
            self.mandatory_headers = []


        # XXX: use proper validation
        assert self.entries_per_page > 0

        self.template = {}
        for name in ("entry", "feed", "index", "path"):
            self.template[name] = parser.get("templates", name)

        # Locate the entries
        self._find_entries()

        # Categorize the entries
        self.tagcloud = TagCloud()

        if not self.entries:
            sys.stderr.write("No articles found!\n")
            sys.exit(0)

        for entry in self.entries:

            sys.stderr.write("ENTRY: %s" % entry.name)
            if len(entry.tags):
                sys.stderr.write(" [ %s ]" % " ".join(entry.tags))
            sys.stderr.write("\n")

            self.tagcloud.add_entry(entry)

        # Assign weight to tags
        self.tagcloud.rank(self.tag_ranks)

    def _find_entries(self):
        for dirpath, dirnames, filenames in os.walk(self.src):

            def _dir_depth(s):
                assert s.startswith(self.src)
                s = s[len(self.src) + 1 :]

                depth = 0
                while len(s):
                    d, f = os.path.split(s)
                    s = d
                    depth += 1

                return depth

            for d in filter(lambda x: x[0] == '.', dirnames):
                dirnames.remove(d)

            if _dir_depth(dirpath) < 1:
                continue

	    # XXX: should be configurable
            if "index.html" not in filenames:
                continue

            self.entries.append(
                BlogEntry(self, dirpath[len(self.src) + 1:])
            )

#############
# Functions #
#############

def usage(progname):
    sys.stdout.write("""WadcomBlog v.%s (r%s)

Usage:

    %s <cfgfile> <dst>
""" % (VERSION, revision, progname))

    sys.exit(1)

def compose_index_first_name(cfg, incomplete_name):
    if cfg.append_first_index_fname_to_links:
        return os.path.join(incomplete_name, cfg.index_first_fname)
    else:
        return incomplete_name

def index_fname(cfg, pagenum):
    if pagenum == 1:
	return cfg.index_first_fname % { "page": "1" }
    else:
	return cfg.index_rest_fname % { "page": pagenum }

def mktmpl_page(cfg, pagenum):
    return _data_obj(
        {
            "link":         index_fname(cfg, pagenum),
            "num":          pagenum,
        }
    )

def mktmpl_tag(cfg, root, tag):
    return _data_obj(
        {
            "index_path":   compose_index_first_name(
                                cfg,
                                "%s/tag/%s" % (root, tag.name)
                            ),
            "name":         tag.name,
            "posts":        len(tag.entries), # XXX: should be len(tag)
            "rank":         tag.rank,
        }
    )

def process_template(cfg, name, data, f):
    lib.ezt.Template(
        os.path.join(cfg.src, cfg.template["path"], name), 
        compress_whitespace=0).generate(f, data)

def install_page(cfg, tpl_name, dst, level, caption, context_maker, feed=None):

    assert os.path.exists(dst)

    assert level >= 0

    if level > 0:
        root = "/".join([ ".." ] * level)
    else:
        root = "."

    pages = context_maker(root)

    pages_tmpl = [mktmpl_page(cfg, x + 1) for x in range(len(pages))]

    page_prev = ""
    for p in range(len(pages)):

        if p == 0:
            fname = index_fname(cfg, 1)
        else:
            fname = index_fname(cfg, p + 1)

        if p == (len(pages) - 1):
            page_next = ""
        else:
            page_next = index_fname(cfg, p + 2)

        f = open(os.path.join(dst, fname), "w")

        data = {
            "blog_title":   cfg.blog_title,
            "caption":      caption,
            "feed":         feed,
            "page_next":    page_next,
            "page_number":  p + 1,
            "page_prev":    page_prev,
            "pages":        pages_tmpl,
            "pages_total":  len(pages),
            "root_path":    root,
            "root_url":     cfg.root_url,
            "tags":         [
                mktmpl_tag(cfg, root, cfg.tagcloud[x]) for x in cfg.tagcloud
            ]
        };

        data.update(pages[p])

        process_template(cfg, tpl_name, data, f)

        f.write("""
<!-- 
    Generated by WadcomBlog v.%s (r%s)

    WadcomBlog is (c) 2006-2008 Vlad Skvortsov, vss@73rus.com

    http://vss.73rus.com/wadcomblog
!-->""" 
            % (VERSION, revision)
        )

        page_prev = fname
    

def prepare_entry(cfg, root, entry):

    # XXX: add weekday 
    t = entry.datetime

    monthname = { 
        "01": "Jan",    
        "02": "Feb",    
        "03": "Mar",    
        "04": "Apr",    
        "05": "May",    
        "06": "Jun",    
        "07": "Jul",    
        "08": "Aug",    
        "09": "Sep",    
        "10": "Oct",    
        "11": "Nov",    
        "12": "Dec" 
    }

    return [ {
        "entries":    _data_obj(
            {
                "body":         entry.body,
                "datetime":     "%s %s %s, %s:%s" % (
                                    monthname[t[4:6]], t[6:8], t[:4], 
                                    t[8:10], t[10:12]
                                ),
                "header":       entry.header,
                "path":         entry.path(root),
                "permalink":    entry.path(cfg.root_url),
                "rfcdatetime":  entry.rfcdatetime,
                "tags":         [ 
				    mktmpl_tag(cfg, root, cfg.tagcloud[x]) 
				    for x in entry.tags
                                ],
                "title":        entry.title,
                "uheader":      entry.user_header,
            }
        )
    } ]

def chrono_sort(entries):
    """
    Returns a list of tuples (timestamp, entry) for all entries in ENTRIES in
    reverse chronological order.
    """

    entries = [ (x.datetime, x) for x in entries ]
    entries.sort()
    entries.reverse()

    return entries

def prepare_index(cfg, root, entries, entries_per_page):

    if entries_per_page is None:
        entries_per_page = len(entries)

    entries = chrono_sort(entries)

    res = []

    start_at = 0
    while start_at < len(entries):
        slice = entries[start_at:(start_at + entries_per_page)]
        res.append({
            "entries":  map(
                lambda (ts, entry): 
                    prepare_entry(cfg, root, entry)[0]["entries"],
                slice
            ),

            "feed_path":	cfg.feed_fname
        })

        start_at += entries_per_page

    return res

def create_feed(cfg, dstdir, title, channel_link, description, entries):

    fname = os.path.join(dstdir, cfg.feed_fname)

    sys.stderr.write("FEED: %s\n" % fname)

    data = {
        "channel":  _data_obj(
            {
                "description":  description,
                "link":         channel_link,
                "title":        title,
            }
        ),
    }

    data.update(prepare_index(cfg, None, entries, cfg.feed_size)[0])

    process_template(cfg, cfg.template["feed"], data, open(fname, "w"))

###############
# ENTRY POINT #
###############

revision = "$Id: wadcomblog.py 1823 2008-07-05 05:27:18Z vss $".split()[2]

if len(sys.argv) != 3:
    usage(sys.argv[0])

cfgfile, dst = sys.argv[1:]

if not os.path.exists(dst):
    sys.stderr.write("%s doesn't exist\n" % dst)
    sys.exit(1)

if not os.path.isdir(dst):
    sys.stderr.write("%s is not a directory\n" % dst)
    sys.exit(1)

cfg = Config(cfgfile)

# Install static content
for f in cfg.static:

    srcname = os.path.join(cfg.src, f)
    dstname = os.path.join(dst, f)

    sys.stderr.write("STATIC: %s\n    %s -> %s\n" % (f, srcname, dstname))

    if not os.path.exists(srcname):
        sys.stderr.write("%s: doesn't exist\n" % srcname)
        sys.exit(1)

    dstdir = os.path.dirname(dstname)

    makedirs(dstdir)

    shutil.copy(srcname, dstname) 

#
# Now write principal articles
#

for entry in cfg.entries:
    sys.stderr.write("\t%s\n" % entry.name)

    entry.install(dst)

#
# Install root page (latest entries)
#
install_page(cfg, cfg.template["index"], dst, 0, cfg.front_page_caption,
    lambda root: prepare_index(cfg, root, cfg.entries, cfg.entries_per_page)
)


create_feed(cfg, dst, cfg.blog_title, cfg.root_url,
    "The latest entries of %s" % cfg.blog_title,
    cfg.entries
)

#
# Install tag indices
#
for tag in cfg.tagcloud:
    sys.stderr.write("TAG: %s\n" % tag)

    tag_dir = os.path.join(dst, "tag", tag)

    makedirs(tag_dir)

    install_page(cfg, cfg.template["index"], tag_dir, 2, "Tag: %s" % tag,
        lambda root: prepare_index(
            cfg, root, cfg.tagcloud[tag].entries, cfg.entries_per_page
        )
    )

    create_feed(cfg, tag_dir,
        "%s: %s" % (cfg.blog_title, tag),
        "%s/tag/%s" % (cfg.root_url, tag),
        "Entries of %s tagged with '%s'" % (cfg.blog_title, tag),
        cfg.tagcloud[tag].entries
    )
