#!/usr/local/bin/python2.6
#
# wadcomblog-mk-redirects - create redirects from entries created with
# WadcomBlog 0.2 to new permanent links. 
#
# Copyright (c) 2006-2008 Vlad Skvortsov, vss@73rus.com
#
# $Id: wblg-mk-redirects.py 1826 2008-07-05 05:32:53Z vss $
# $HeadURL: https://svn.xiolabs.com/vss/wadcomblog/tags/0.3/src/wblg-mk-redirects.py $
#
# Redistribution and use in source and binary forms, with or without 
# modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, 
# this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright notice, 
# this list of conditions and the following disclaimer in the documentation 
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#

import ConfigParser
import email
import errno
import os
import shutil
import sys

import lib.ezt

DIRECTIVE = "<!-- WADCOMBLOG:"
DIRECTIVE_HEADER_END = DIRECTIVE + " header-end"

class _data_obj:
    def __init__(self, d):
        vars(self).update(d)

class BlogEntry:
    def __init__(self, root_url, srcdir, fname):

        assert root_url

        s = os.path.split(fname)
        
        assert len(s) == 2

        self.category, self.name = s

	# XXX: source file name hardcoded as "index.html"
        name = os.path.join(srcdir, fname, "index.html")
        data = email.message_from_file(open(name))

        if "Title" not in data: 
            sys.stderr.write("%s: no 'Title' header\n" % name)
            raise ValueError

        self.title = data["Title"]

        if "Timestamp" not in data: 
            sys.stderr.write("%s: no 'Timestamp' header\n" % name)
            raise ValueError

        self.datetime = data["Timestamp"]

        tags = []

        if "Tags" in data: 
            tags = data["Tags"].strip().split()
            tags.sort()

        self.tags = tags

        self.header = self.body = None

        self.permalink = os.path.join(root_url,
            self.datetime[0:4], self.datetime[4:6], self.datetime[6:8],
            self.name
        )

        contents = ""
        for str in data.get_payload().split("\n"):
            
            if str.strip().startswith(DIRECTIVE_HEADER_END):

                if self.header is None:

                    self.header = contents
                    contents = ""

                else:
                    sys.stderr.write(
                        "%s: duplicate 'header-end' directive, ignored\n"
                        % name
                    )

            else:
                contents += "\n" + str

        if self.header is None:   
            sys.stderr.write("%s: no 'header-end' directive\n" % name)
            raise ValueError

        self.body = contents

class Tag:
    def __init__(self, name, entry):
        self.entries = [ entry ]
        self.name = name

    def add_entry(self, entry):
        self.entries.append(entry)

class TagCloud:
    def __getitem__(self, key):
        return self.tags[key]

    def __init__(self):
        self.tags = {}

    def __iter__(self):
        tags = self.tags.keys()
        tags.sort()
        return tags.__iter__()

    def add_entry(self, entry):
        for tag in entry.tags:
            if tag in self.tags:
                self.tags[tag].add_entry(entry)
            else:
                self.tags[tag] = Tag(tag, entry)

    def rank(self, maxrating):
        """
        Go over all exisiting tags and assign rating from 1 (most popular)
        to MAXRATING (least popular).
        """

        assert maxrating >= 1

        lens = {}
        for l in map(lambda x: len(self.tags[x].entries), self.tags.keys()):
            lens[l] = True

        lens = lens.keys()
        lens.sort()

        # Number of different entry-per-tag counts
        weights = len(lens)

        if weights < maxrating:
            maxrating = weights
            assert maxrating >= 1

        step = float(weights) / maxrating

        weight = {}
        next = step
        rank = maxrating
        for i in range(weights):
            if i >= next:
                next += step
                rank -= 1

            weight[lens[i]] = rank

        for tag in self.tags.keys():
            self.tags[tag].rank = weight[len(self.tags[tag].entries)]

class Config:
    def __init__(self, fname):

        s = os.path.split(fname)
        if s[0]:
            self.src = s[0]
        else:
            self.src = "."

        parser = ConfigParser.ConfigParser()
        parser.read(fname)

        self.template = self.root_url = self.blog_title = self.feed_fname = \
            self.feed_size = None

        self.entries = []
        self.category_title = {}

        # Fetch data from the configuration
        self.blog_title = parser.get("general", "blog-title")
        self.front_page_caption = parser.get("general", "front-page-caption")
        self.entries_per_page = int(parser.get("general", "entries-per-page"))
        self.feed_fname = parser.get("general", "feed")
        self.feed_size = int(parser.get("general", "feed-size"))
        self.index_first_fname = parser.get(
	    "general", "index-first-fname", raw=True
	)
        self.index_rest_fname = parser.get(
	    "general", "index-rest-fname", raw=True
	)

        try:
            yesno = parser.get(
                "general", 
                "append-first-index-fname-to-links"
            ).lower()
        except ConfigParser.NoOptionError:
            yesno = "no"

        self.append_first_index_fname_to_links = (yesno == "yes")

        self.root_url = parser.get("general", "root")
        self.static = parser.get("general", "static").split()
        self.tag_ranks = int(parser.get("general", "tag-ranks"))

        if self.feed_size < 1:
            sys.stderr.write(
                "Invalid general/feed-size value: %s\n" % self.feed_size
            )
            sys.exit(1)

        # XXX: use proper validation
        assert self.entries_per_page > 0

        self.template = {}
        for name in ("entry", "feed", "index", "path"):
            self.template[name] = parser.get("templates", name)

        for name, value in parser.items("category-titles"):
            self.category_title[name] = value

        # Locate the entries
        self._find_entries()

        # Categorize the entries
        self.category = {}
        self.tagcloud = TagCloud()

        for entry in self.entries:

            sys.stderr.write("ENTRY: %s in %s" % (entry.name, entry.category))
            if len(entry.tags):
                sys.stderr.write(" [ %s ]" % " ".join(entry.tags))
            sys.stderr.write("\n")

            if entry.category in self.category:
                self.category[entry.category].append(entry)
            else:
                self.category[entry.category] = [ entry ]

            self.tagcloud.add_entry(entry)

        # Assign weight to tags
        self.tagcloud.rank(self.tag_ranks)

        # Check/initialize category titles
        for c in self.category_title:
            if c not in self.category:
                sys.stderr.write("%s: category '%s' is not defined\n" % 
                    (fname, c)
                )
                sys.exit(1)

        for c in self.category:
            if c not in self.category_title:
                self.category_title[c] = c.capitalize()

    def _find_entries(self):
        for dirpath, dirnames, filenames in os.walk(self.src):

            def _dir_depth(s):
                assert s.startswith(self.src)
                s = s[len(self.src) + 1 :]

                depth = 0
                while len(s):
                    d, f = os.path.split(s)
                    s = d
                    depth += 1

                return depth

            for d in filter(lambda x: x[0] == '.', dirnames):
                dirnames.remove(d)

            if _dir_depth(dirpath) != 2:
                continue

	    # XXX: should be configurable
            if "index.html" not in filenames:
                continue

            self.entries.append(
                BlogEntry(self.root_url, self.src, dirpath[len(self.src) + 1:])
            )

#############
# Functions #
#############

def usage(progname):
    sys.stdout.write("""

Usage:

    %s <cfgfile> <redirect_fname> <redirect_template> <dst>

The utility will use <cfggile> to create files names <redirect_fname> (e.g.
'.htaccess') in directories corresponding to WadcomBlog 0.2 content layout.
The file contents are based on <redirect_template>. <dst> is the output root.

The list of created files (absolute paths) is printed to the standard output.

A <redirect_template> for Apache (.htacess files) might look like:

    Redirect /where-my-blog-is/[path] [permalink]

([path] is the 'old' path to the article, relative to the blog root;
[permalink] is the new permalink)

""" % progname)

    sys.exit(1)

def install_page(cfg, fname, tpl_name, dst, entry_path, permalink, level):

    assert os.path.exists(dst)

    assert level >= 0

    if level > 0:
        root = "/".join([ ".." ] * level)
    else:
        root = "."

    f = open(os.path.join(dst, fname), "w")

    data = { "path": entry_path, "permalink": permalink }
    lib.ezt.Template(tpl_name,
        compress_whitespace=0).generate(f, data)

    sys.stdout.write("%s\n" % os.path.join(dst, fname))


###############
# ENTRY POINT #
###############

if len(sys.argv) != 5:
    usage(sys.argv[0])

cfgfile, redirect_fname, redirect_tmpl, dst = sys.argv[1:]

if not os.path.exists(dst):
    sys.stderr.write("%s doesn't exist\n" % dst)
    sys.exit(1)

if not os.path.isdir(dst):
    sys.stderr.write("%s is not a directory\n" % dst)
    sys.exit(1)

cfg = Config(cfgfile)

# Install static content
for f in cfg.static:

    srcname = os.path.join(cfg.src, f)
    dstname = os.path.join(dst, f)

    sys.stderr.write(
"""STATIC: %s
    %s -> %s\n"""
        % (f, srcname, dstname))

    if not os.path.exists(srcname):
        sys.stderr.write("%s: doesn't exist\n" % srcname)
        sys.exit(1)

    dstdir = os.path.dirname(dstname)

    try:
        os.makedirs(dstdir)
    except OSError, (code, str):
        if code != errno.EEXIST:
            raise

    shutil.copy(srcname, dstname) 

#
# Now write principal articles
#
for category in cfg.category:
    sys.stderr.write("CATEGORY: %s\n" % category)

    dstdir = os.path.join(dst, category)

    if not os.path.exists(dstdir):
        os.mkdir(dstdir)

    #
    # Install entries in this category
    #
    for entry in cfg.category[category]:
        sys.stderr.write("\t%s\n" % entry.name)

        entry_dir = os.path.join(dstdir, entry.name)

        if not os.path.exists(entry_dir):
            os.mkdir(entry_dir)

        install_page(cfg, redirect_fname, redirect_tmpl, entry_dir,
            os.path.join(category, entry.name),
            entry.permalink,
            2)
